package gonxlog

/*
#cgo linux LDFLAGS: -Wl,--unresolved-symbols=ignore-all
#cgo windows LDFLAGS: -Wl,--unresolved-symbols=ignore-all
#cgo darwin LDFLAGS: -Wl,-undefined,suppress -flat_namespace
#include <stdlib.h>
#include <stdint.h>

struct nx_module_t;
typedef struct nx_module_t nx_module_t;

struct nx_record_t;
typedef struct nx_record_t nx_record_t;

struct nx_value_t;
typedef struct nx_value_t nx_value_t;

typedef struct nx_go_module_t
{
    nx_module_t *module;
} nx_go_module_t;


typedef struct nx_go_logdata_t
{
    nx_record_t *record;
    int	  external;
} nx_go_logdata_t;


typedef struct nx_go_ctx_t
{
    nx_go_module_t module;
    nx_go_logdata_t logdata;
    nx_value_t **argv;
    int argc;

} nx_go_ctx_t;

int ctx_get_argc(nx_go_ctx_t *ctx)
{
	return ctx->argc;
}

nx_value_t * ctx_get_argv(nx_go_ctx_t *ctx, int i)
{
	return ctx->argv[i];
}

int ctx_has_logdata(nx_go_ctx_t *ctx)
{
	return ctx->logdata.record != NULL;
}

int ctx_has_module(nx_go_ctx_t *ctx)
{
	return ctx->module.module != NULL;
}

nx_go_module_t * ctx_get_module(nx_go_ctx_t *ctx)
{
	return &ctx->module;
}

nx_go_logdata_t * ctx_get_logdata(nx_go_ctx_t *ctx)
{
	return &ctx->logdata;
}

typedef struct
{
    char ** elts;
    int nelts;
} nx_go_str_array_t;



nx_go_logdata_t *nx_go_logdata_new();
void nx_go_logdata_free(nx_go_logdata_t *ld);
void nx_go_logdata_post(nx_go_module_t *module, nx_go_logdata_t *data);


char   *nx_go_logdata_field_get_string(nx_go_logdata_t *ld, const char *field);
int64_t nx_go_logdata_field_get_int(nx_go_logdata_t *ld, const char *field);
int     nx_go_logdata_field_get_boolean(nx_go_logdata_t *ld, const char *field);
void    nx_go_logdata_field_set_string(nx_go_logdata_t *ld, const char *field, const char *value);
void    nx_go_logdata_field_set_int(nx_go_logdata_t *ld, const char *field, int64_t value);
void    nx_go_logdata_field_set_boolean(nx_go_logdata_t *ld, const char *field, int value);
void    nx_go_logdata_field_delete(nx_go_logdata_t *ld, const char *field);
int     nx_go_logdata_field_exists(nx_go_logdata_t *ld, const char *field);
int     nx_go_logdata_field_get_type(nx_go_logdata_t *ld, const char *field);


int	nx_go_value_get_type(nx_value_t *value);
int64_t	nx_go_value_get_int(nx_value_t *value);
char	*nx_go_value_get_string(nx_value_t *value);
unsigned int	nx_go_value_get_boolean(nx_value_t *value);
int     nx_go_value_is_defined(nx_value_t *value);

nx_go_str_array_t *nx_go_logdata_get_fieldnames(nx_go_logdata_t *ld);
void nx_go_array_free(nx_go_str_array_t *array);

void nx_go_log_info(const char *what);
void nx_go_log_debug(const char *what);
void nx_go_log_warning(const char *what);
void nx_go_log_error(const char *what);

void nx_go_event_add(nx_go_module_t *module, int delay_msec);

*/
import "C"

import (
	"fmt"
	"unsafe"
	"runtime"
)

type NxFieldType int

const (
	NxFieldTypeInt int = 1
	NXFieldTypeString int = 2
	NxFieldTypeBoolean int = 5
)

// Module wrapper for nxlog api
type nxModule struct {
	module *C.nx_go_module_t
}

// Wrapper for nxlog 'logdata' structure
type nxLogdata struct {
	ld *C.nx_go_logdata_t
}

// Conversion to Golang wrapper structure from nxlog api module struct
func nxModuleConv(module unsafe.Pointer) *nxModule {
	return &nxModule{module: (*C.nx_go_module_t)(module)}
}

// Conversion to Golang wrapper structure from nxlog api logdata struct
func nxLogdataConv(ld unsafe.Pointer) *nxLogdata {
	return &nxLogdata{ld: (*C.nx_go_logdata_t)(ld)}
}


func GetModule(ctx unsafe.Pointer) (*nxModule, bool) {
	ptr := (*C.nx_go_ctx_t)(ctx);
	if C.ctx_has_module(ptr) != 0 {
		return nxModuleConv(unsafe.Pointer(C.ctx_get_module(ptr))), true;
	}
	return nil, false
}

func GetLogdata(ctx unsafe.Pointer) (*nxLogdata, bool) {
	ptr := (*C.nx_go_ctx_t)(ctx);
	if C.ctx_has_logdata(ptr) != 0 {
		return nxLogdataConv(unsafe.Pointer(C.ctx_get_logdata(ptr))), true;
	}
	return nil, false
}


func nxValue(val unsafe.Pointer) (interface{}, bool) {

	ptr := (*C.nx_value_t)(val)
	if C.nx_go_value_is_defined(ptr) == 0 {
		return nil, false
	}
	switch C.nx_go_value_get_type(ptr) {

	case 1:
		return C.nx_go_value_get_int(ptr), true

	case 5:
		return C.nx_go_value_get_boolean(ptr) != 0, true

	default:
		cVal := C.nx_go_value_get_string(ptr)
		defer C.free(unsafe.Pointer(cVal))
		return  C.GoString(cVal), true
	}
	return nil, false
}

func GetArg(ctx unsafe.Pointer, i int) (interface{}, bool) {
	argc := GetArgc(ctx);
	ptr := (*C.nx_go_ctx_t)(ctx);
	if i >= argc || i < 0 {
		return nil, false
	}
	return nxValue(unsafe.Pointer(C.ctx_get_argv(ptr, (C.int)(i))))
}

func GetArgc(ctx unsafe.Pointer) int {
	ptr := (*C.nx_go_ctx_t)(ctx);
	return (int)(C.ctx_get_argc(ptr));
}

// GoString converts C-array of strings to Golang string slice
func GoStrings(argc C.int, argv **C.char) []string {

	length := int(argc)
	tmpslice := (*[1 << 30]*C.char)(unsafe.Pointer(argv))[:length:length]
	gostrings := make([]string, length)
	for i, s := range tmpslice {
		gostrings[i] = C.GoString(s)
	}
	return gostrings
}


// NxLogdataFree is destroing the logdata structure
// and freeing memory
func nxLogdataFree(ld *nxLogdata) {
	C.nx_go_logdata_free(ld.ld)
}

// Return string representation from the
// logdata field
func (ld nxLogdata) Get(field string) (interface{}, bool) {

	cName := C.CString(field)
	defer C.free(unsafe.Pointer(cName))

	if C.nx_go_logdata_field_exists(ld.ld, cName) == 0 {
		return nil, false
	}

	if C.nx_go_logdata_field_get_type(ld.ld, cName) == 1 {
		return C.nx_go_logdata_field_get_int(ld.ld, cName), true
	}

	if C.nx_go_logdata_field_get_type(ld.ld, cName) == 5 {
		return C.nx_go_logdata_field_get_boolean(ld.ld, cName) != 0, true
	}

	cVal := C.nx_go_logdata_field_get_string(ld.ld, cName)
	defer C.free(unsafe.Pointer(cVal))
	return  C.GoString(cVal), true
}

func (ld nxLogdata) GetString(field string) (string, bool) {
	cName := C.CString(field)
	defer C.free(unsafe.Pointer(cName))
	cVal := C.nx_go_logdata_field_get_string(ld.ld, cName)
	defer C.free(unsafe.Pointer(cVal))
	return  C.GoString(cVal), true
}

// nxlogdata.Set() sets logdata field value
func (ld nxLogdata) Set(field string, val interface{}) {

	cField := C.CString(field)
	defer C.free(unsafe.Pointer(cField))

	switch i := val.(type) {
	case nil:
	case int:
		C.nx_go_logdata_field_set_int(ld.ld, cField, C.int64_t(i))
	case int64:
		C.nx_go_logdata_field_set_int(ld.ld, cField, C.int64_t(i))
	case string:
		cVal := C.CString(i)
		defer C.free(unsafe.Pointer(cVal))
		C.nx_go_logdata_field_set_string(ld.ld, cField, cVal)
	case bool:
		if i {
			C.nx_go_logdata_field_set_boolean(ld.ld, cField, 1)
		} else {
			C.nx_go_logdata_field_set_boolean(ld.ld, cField, 0)
		}
	default:
		cVal := C.CString(fmt.Sprint(val))
		defer C.free(unsafe.Pointer(cVal))
		C.nx_go_logdata_field_set_string(ld.ld, cField, cVal)
	}
}

// nxlogdata.Delete() remove field from logdata
func (ld nxLogdata) Delete(field string) {
	cField := C.CString(field)
	defer C.free(unsafe.Pointer(cField))
	C.nx_go_logdata_field_delete(ld.ld, cField)
}

// nxlogdata.Fields() returns slice of field names
func (ld nxLogdata) Fields() []string {
	cFields := C.nx_go_logdata_get_fieldnames(ld.ld)
	defer C.nx_go_array_free(cFields)
	return GoStrings(cFields.nelts, cFields.elts)
}


func (mod nxModule) NxLogdataNew() *nxLogdata {
	rv := &nxLogdata{ld: C.nx_go_logdata_new()}
	runtime.SetFinalizer(rv, func ( ld *nxLogdata ) {nxLogdataFree(ld)})
	return rv
}


// nxmodule.Post() will put logdata struct to further processing
func (mod nxModule) Post(ld *nxLogdata) {
	C.nx_go_logdata_post(mod.module, ld.ld)
}

// Adds READ event to nxlog. This allows to call READ in the future
func (mod nxModule) AddEvent() {
	C.nx_go_event_add(mod.module, 0)
}

func (mod nxModule) AddEventDelayed(mSec C.int) {
	C.nx_go_event_add(mod.module, mSec)
}

// Writes INFO string into nxlog log file
func LogInfo(what string){
	cWhat := C.CString(what)
	defer C.free(unsafe.Pointer(cWhat))
	C.nx_go_log_info(cWhat)
}

// Writes DEBUG string into nxlog log file
func LogDebug(what string) {
	cWhat := C.CString(what)
	defer C.free(unsafe.Pointer(cWhat))
	C.nx_go_log_debug(cWhat)
}

// Writes WARNING string into nxlog log file
func LogWarn(what string) {
	cWhat := C.CString(what)
	defer C.free(unsafe.Pointer(cWhat))
	C.nx_go_log_warning(cWhat)
}

// Writes ERROR string into nxlog log file
func LogError(what string) {
	cWhat := C.CString(what)
	defer C.free(unsafe.Pointer(cWhat))
	C.nx_go_log_error(cWhat)
}
